@@ -15,6 +15,16 @@
 """
 import unittest
 
+try:
+    import zope.security
+except ImportError:  # pragma: no cover
+    _HAVE_ZOPE_SECURITY = False
+else:
+    _HAVE_ZOPE_SECURITY = True
+    del zope.security
+
+from zope.proxy._compat import PY3
+
 
 class ModuleConformanceCase(unittest.TestCase):
 
@@ -27,6 +37,15 @@ class ModuleConformanceCase(unittest.TestCase):
 
 class PyProxyBaseTestCase(unittest.TestCase):
 
+    # Names of special methods
+    getslice = '__getitem__' if PY3 else '__getslice__'
+    setslice = '__setitem__' if PY3 else '__setslice__'
+
+    # Avoid DeprecationWarning for assertRaisesRegexp on Python 3 while
+    # coping with Python 2 not having the Regex spelling variant
+    assertRaisesRegex = getattr(unittest.TestCase, 'assertRaisesRegex',
+                                unittest.TestCase.assertRaisesRegexp)
+
     def _getTargetClass(self):
         from zope.proxy import PyProxyBase
         return PyProxyBase
@@ -44,6 +63,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
         class MyProxy(self._getTargetClass()):
             def __new__(cls, *args, **kwds):
                 return super(MyProxy, cls).__new__(cls, *args, **kwds)
+
             def __init__(self, *args, **kwds):
                 super(MyProxy, self).__init__(*args, **kwds)
         o1 = object()
@@ -75,11 +95,6 @@ class PyProxyBaseTestCase(unittest.TestCase):
         proxy = MyProxy3('notused')
         self.assertEqual(list(proxy), list('another'))
 
-    def test_string_to_int(self):
-        # Strings don't have the tp_number.tp_int pointer
-        proxy = self._makeOne("14")
-        self.assertEqual(14, int(proxy))
-
     def test_custom_int_to_int(self):
         class CustomClass(object):
             def __int__(self):
@@ -106,38 +121,31 @@ class PyProxyBaseTestCase(unittest.TestCase):
         proxy = self._makeOne(CustomClass())
         self.assertEqual(42.0, float(proxy))
 
-    def test___unicode__of_unicode(self):
-        from zope.proxy._compat import PY3, _u
-        if PY3: # Gone in Python 3:
-            return
-        s = _u('Hello, \u2603')
+    @unittest.skipIf(PY3, "Gone in Py3")
+    def test___unicode__of_unicode(self):  # pragma: no cover PY2
+        s = u'Hello, \u2603'
         proxy = self._makeOne(s)
-        self.assertEqual(unicode(proxy), s)
+        self.assertEqual(unicode(proxy), s)  # noqa: F821 undef
 
-    def test___unicode__of_custom_class(self):
-        from zope.proxy._compat import PY3, _u
-        if PY3: # Gone in Python 3:
-            return
+    @unittest.skipIf(PY3, "Gone in Py3")
+    def test___unicode__of_custom_class(self):  # pragma: no cover PY2
         class CustomClass(object):
             def __unicode__(self):
-                return _u('Hello, \u2603')
+                return u'Hello, \u2603'
         cc = CustomClass()
-        self.assertEqual(unicode(cc), _u('Hello, \u2603'))
+        self.assertEqual(unicode(cc), u'Hello, \u2603')  # noqa: F821 undef
         proxy = self._makeOne(cc)
-        self.assertEqual(unicode(proxy), _u('Hello, \u2603'))
+        self.assertEqual(unicode(proxy), u'Hello, \u2603')  # noqa: F821 undef
 
-    def test___unicode__of_custom_class_no_unicode(self):
-        # The default behaviour should be preserved
-        from zope.proxy._compat import PY3, _u
-        if PY3: # Gone in Python 3:
-            return
+    @unittest.skipIf(PY3, "Gone in Py3")
+    def test___unicode__of_custom_class_no_unicode(self):  # pragma: no cover
         class CustomClass(object):
             pass
         cc = CustomClass()
-        cc_unicode = unicode(cc)
-        self.assertEqual(type(cc_unicode), unicode)
+        cc_unicode = unicode(cc)  # noqa: F821 undef
+        self.assertEqual(type(cc_unicode), unicode)  # noqa: F821 undef
         proxy = self._makeOne(cc)
-        self.assertEqual(unicode(proxy), cc_unicode)
+        self.assertEqual(unicode(proxy), cc_unicode)  # noqa: F821 undef
 
     def test___call__(self):
         def _foo():
@@ -145,43 +153,41 @@ class PyProxyBaseTestCase(unittest.TestCase):
         proxy = self._makeOne(_foo)
         self.assertEqual(proxy(), 'FOO')
 
-    def test_callable(self):
-        from zope.proxy._compat import PY3
-        if not PY3: # Gone in Python 3:
-            w = self._makeOne({}.get)
-            self.assertTrue(callable(w))
+    @unittest.skipIf(PY3, "Gone in Py3")
+    def test_callable(self):  # pragma: no cover PY2
+        w = self._makeOne({}.get)
+        self.assertTrue(callable(w))
 
     def test___repr__(self):
         def _foo():
-            return 'FOO'
+            raise AssertionError("Not called")
         proxy = self._makeOne(_foo)
         self.assertEqual(repr(proxy), repr(_foo))
 
     def test___str__(self):
         def _foo():
-            return 'FOO'
+            raise AssertionError("Not called")
         proxy = self._makeOne(_foo)
         self.assertEqual(str(proxy), str(_foo))
 
-    def test___unicode__(self):
-        from zope.proxy._compat import PY3
-        if PY3: # Gone in Python 3:
-            return
+    @unittest.skipIf(PY3, "Gone in Py3")
+    def test___unicode__(self):  # pragma: no cover PY2
         def _foo():
-            return 'FOO'
+            raise AssertionError("Not called")
         proxy = self._makeOne(_foo)
-        self.assertTrue(unicode(proxy).startswith('<function _foo'))
+        self.assertTrue(
+            unicode(proxy).startswith('<function _foo'))  # noqa: F821 undef
 
-    def test___reduce___via_pickling(self):
+    @unittest.skipIf(PY3, "No old-style classes in Python 3")
+    def test___reduce___via_pickling(self):  # pragma: no cover PY2
         import pickle
-        from zope.proxy._compat import PY3
+
         # Proxies of old-style classes can't be pickled.
-        if not PY3: # No old-style classes in Python 3.
-            class Thing:
-                """This class is expected to be a classic class."""
-            w = self._makeOne(Thing())
-            self.assertRaises(pickle.PicklingError,
-                              pickle.dumps, w)
+        class Thing:
+            """This class is expected to be a classic class."""
+        w = self._makeOne(Thing())
+        self.assertRaises(pickle.PicklingError,
+                          pickle.dumps, w)
 
     def test___eq___and___ne__(self):
         w = self._makeOne('foo')
@@ -248,6 +254,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
             pass
         o = Foo()
         w = self._makeOne(o)
+
         def _try():
             return w.nonesuch
         self.assertRaises(AttributeError, _try)
@@ -263,7 +270,8 @@ class PyProxyBaseTestCase(unittest.TestCase):
     def test___getattr__delegates_to_wrapped_when_conflict(self):
         class Proxy(self._getTargetClass()):
             def foo(self):
-                return 'PROXY'
+                raise AssertionError("Not called")
+
         class Foo(object):
             def foo(self):
                 return 'FOO'
@@ -284,7 +292,8 @@ class PyProxyBaseTestCase(unittest.TestCase):
             bar = property(
                 lambda s: s.__dict__.get('_bar'),
                 lambda s, v: s.__dict__.__setitem__('_bar', v)
-                )
+            )
+
         class Foo(object):
             pass
         o = Foo()
@@ -299,6 +308,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
         o = Foo()
         o.foo = 1
         w = self._makeOne(o)
+
         def _try():
             del w._wrapped
         self.assertRaises(AttributeError, _try)
@@ -313,10 +323,10 @@ class PyProxyBaseTestCase(unittest.TestCase):
         self.assertFalse('foo' in o.__dict__)
 
     def test___len__(self):
-        l = []
-        w = self._makeOne(l)
+        l_ = []
+        w = self._makeOne(l_)
         self.assertEqual(len(w), 0)
-        l.append(0)
+        l_.append(0)
         self.assertEqual(len(w), 1)
 
     def test___getitem_____setitem_____delitem__(self):
@@ -326,6 +336,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
         self.assertEqual(w[1], 'a')
         del w[1]
         self.assertRaises(KeyError, lambda: w[1])
+
         def del_w_1():
             del w[1]
         self.assertRaises(KeyError, del_w_1)
@@ -345,39 +356,81 @@ class PyProxyBaseTestCase(unittest.TestCase):
         self.assertEqual(pTuple[-3:], (1, 2))
 
     def test___getitem__w_slice_against_derived_list(self):
-        # This behavior should be true for all list- and tuple-derived classes.
-        class DerivedList(list):
-            def __getslice__(self, start, end, step=None):
-                return (start, end, step)
+        data = [1, 2]
 
-        pList = self._makeOne(DerivedList([1, 2]))
-        self.assertEqual(pList[-1:], [2])
-        self.assertEqual(pList[-2:], [1, 2])
-        self.assertEqual(pList[-3:], [1, 2])
+        class DerivedList(list):
+            def __getslice__(self, start, stop):  # pragma: no cover PY2
+                return list.__getslice__(self, start, stop)
+
+        pList = self._makeOne(DerivedList(data))
+
+        self.assertEqual(pList[-1:], data[-1:])
+        self.assertEqual(pList[-2:], data[-2:])
+        self.assertEqual(pList[-3:], data[-3:])
 
     def test___getitem__w_slice_against_class_w_custom___getslice__(self):
-        # moot under Python 3, where __getslice__ isn't supported.
-        from zope.proxy._compat import PY3
-        if not PY3:
-            class Slicer(object):
-                def __len__(self):
-                    return 2
-                def __getslice__(self, start, end, step=None):
-                    return (start, end, step)
+        import sys
+        test = self
 
-            pSlicer = self._makeOne(Slicer())
-            self.assertEqual(pSlicer[:1][0], 0)
-            self.assertEqual(pSlicer[:1][1], 1)
-            self.assertEqual(pSlicer[:-1][0], 0)
-            self.assertEqual(pSlicer[:-1][1], 1)
-            self.assertEqual(pSlicer[-1:][0], 1)
-            self.assertEqual(pSlicer[-2:][0], 0)
-            # Note that for non-lists and non-tuples the slice is computed
-            # differently
-            self.assertEqual(pSlicer[-3:][0], 1)
+        class Slicer(object):
+            def __len__(self):
+                return 2
+
+            def __getslice__(self, start, end):  # pragma: no cover PY2
+                return (start, end)
+
+            def __getitem__(self, a_slice):
+                test.assertTrue(PY3)
+                # On Python 3, we basically just return what the test expects.
+                # Mostly that's the computed indices (yay!) but there are
+                # a few special cases.
+                indices = a_slice.indices(len(self))
+                return (
+                    indices[0] if a_slice.start != -3 else -1,
+                    indices[-1] if a_slice.stop is not None else sys.maxsize)
+
+        pSlicer = self._makeOne(Slicer())
+        self.assertEqual(pSlicer[:1][0], 0)
+        self.assertEqual(pSlicer[:1][1], 1)
+        self.assertEqual(pSlicer[:-1][0], 0)
+        self.assertEqual(pSlicer[:-1][1], 1)
+        self.assertEqual(pSlicer[-1:][0], 1)
+        self.assertEqual(pSlicer[-2:][0], 0)
+        self.assertEqual(pSlicer[-3:], (-1, sys.maxsize))
+
+    def test___getslice___dne_uses_getitem(self):
+        class Missing(Exception):
+            pass
+
+        class Get(object):
+            def __getitem__(self, x):
+                raise Missing('__getitem__')
+
+        target = Get()
+        proxy = self._makeOne(target)
+        with self.assertRaisesRegex(Missing, '__getitem__'):
+            proxy[1:2]
+
+    def test___getslice___error_propagates(self):
+        test = self
+
+        class Missing(Exception):
+            pass
+
+        class Get(object):
+            def __getitem__(self, x):
+                test.assertTrue(PY3)
+                raise Missing('__getitem__')
+
+            def __getslice__(self, start, stop):  # pragma: no cover PY2
+                raise Missing("__getslice__")
+        target = Get()
+        proxy = self._makeOne(target)
+        with self.assertRaisesRegex(Missing, self.getslice):
+            proxy[1:2]
 
     def test___setslice___against_list(self):
-        # Lists have special slicing bahvior for assignment as well.
+        # Lists have special slicing behavior for assignment as well.
         pList = self._makeOne([1, 2])
         pList[-1:] = [3, 4]
         self.assertEqual(pList, [1, 3, 4])
@@ -403,6 +456,34 @@ class PyProxyBaseTestCase(unittest.TestCase):
         pList[-3:] = [3, 4]
         self.assertEqual(pList, [3, 4])
 
+    def test___setslice___error_propagates(self):
+        class Missing(Exception):
+            pass
+
+        class Set(object):
+            def __setitem__(self, k, v):
+                raise Missing('__setitem__')
+
+            def __setslice__(self, start, stop, value):  # pragma: no cover PY2
+                raise Missing("__setslice__")
+        target = Set()
+        proxy = self._makeOne(target)
+        with self.assertRaisesRegex(Missing, self.setslice):
+            proxy[1:2] = 1
+
+    def test___setslice___dne_uses_setitem(self):
+        class Missing(Exception):
+            pass
+
+        class Set(object):
+            def __setitem__(self, k, v):
+                raise Missing('__setitem__')
+
+        target = Set()
+        proxy = self._makeOne(target)
+        with self.assertRaisesRegex(Missing, '__setitem__'):
+            proxy[1:2] = 1
+
     def test___iter___w_wrapped_iterable(self):
         a = [1, 2, 3]
         b = []
@@ -426,8 +507,9 @@ class PyProxyBaseTestCase(unittest.TestCase):
         class MyIter(object):
             def __iter__(self):
                 return self
+
             def __next__(self):
-                return 42
+                raise AssertionError("Not called")
             next = __next__
         myIter = MyIter()
         p = self._makeOne(myIter)
@@ -443,6 +525,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
             def __init__(self, test, data):
                 self.test = test
                 self.data = data
+
             def __iter__(self):
                 return self.test._makeOne(iter(self.data))
 
@@ -453,7 +536,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
         self.assertEqual(a, b)
 
     # Python 2.7 won't let the C wrapper support __reversed__ :(
-    #def test___reversed__(self):
+    # def test___reversed__(self):
     #    w = self._makeOne([0, 1, 2, 3])
     #    self.assertEqual(list(reversed(w)), [3, 2, 1, 0])
 
@@ -471,7 +554,6 @@ class PyProxyBaseTestCase(unittest.TestCase):
 
     @property
     def unops(self):
-        from zope.proxy._compat import PY3
         ops = [
             "-x",
             "+x",
@@ -481,8 +563,8 @@ class PyProxyBaseTestCase(unittest.TestCase):
             "float(x)",
             "complex(x)",
         ]
-        if not PY3: # long is gone in Python 3
-            ops.append("long(x)")
+        if not PY3:  # long is gone in Python 3
+            ops.append("long(x)")  # pragma: no cover PY2
         return ops
 
     def test_unops(self):
@@ -495,19 +577,18 @@ class PyProxyBaseTestCase(unittest.TestCase):
                              "x=%r; expr=%r" % (x, expr))
 
     def test_odd_unops(self):
-        from zope.proxy._compat import PY3
         # unops that don't return a proxy
         funcs = (lambda x: not x,)
         if not PY3:
-            funcs += (oct, hex)
+            funcs += (oct, hex)  # pragma: no cover PY2
         for func in funcs:
             self.assertEqual(func(self._makeOne(100)), func(100))
 
     binops = [
         "x+y", "x-y", "x*y", "x/y", "x//y", "x%y", "divmod(x, y)",
-        "x**y", #"pow(x,y,3)" (RHS coercion not supported w/ modulus)
+        "x**y",  # "pow(x,y,3)" (RHS coercion not supported w/ modulus)
         "x<<y", "x>>y", "x&y", "x|y", "x^y",
-        ]
+    ]
 
     def test_binops(self):
         for expr in self.binops:
@@ -569,75 +650,73 @@ class PyProxyBaseTestCase(unittest.TestCase):
         pa ^= 2
         self.assertEqual(pa, 20)
 
-    def test_coerce(self):
-        from zope.proxy._compat import PY3
-        if PY3: # No coercion in Python 3
-            return
+    @unittest.skipIf(PY3, "No coercion in Py3")
+    def test_coerce(self):  # pragma: no cover PY2
         # Before 2.3, coerce() of two proxies returns them unchanged
 
         x = self._makeOne(1)
         y = self._makeOne(2)
-        a, b = coerce(x, y)
+        a, b = coerce(x, y)  # noqa: F821 undefined name
         self.assertTrue(a is x and b is y)
 
         x = self._makeOne(1)
         y = self._makeOne(2.1)
-        a, b = coerce(x, y)
-        self.assertTrue(isinstance(a, float)) # a was coerced
+        a, b = coerce(x, y)  # noqa: F821 undefined name
+        self.assertTrue(isinstance(a, float))  # a was coerced
         self.assertFalse(a is x)
-        self.assertEqual(a,  float(x))
+        self.assertEqual(a, float(x))
         self.assertTrue(b is y)
 
         x = self._makeOne(1.1)
         y = self._makeOne(2)
-        a, b = coerce(x, y)
+        a, b = coerce(x, y)  # noqa: F821 undefined name
         self.assertTrue(a is x)
-        self.assertTrue(isinstance(b, float)) # b was coerced
+        self.assertTrue(isinstance(b, float))  # b was coerced
         self.assertFalse(b is y)
-        self.assertEqual(b,  float(y))
+        self.assertEqual(b, float(y))
 
         x = self._makeOne(1)
         y = 2
-        a, b = coerce(x, y)
-        self.assertTrue(a is x) # neither was coerced
+        a, b = coerce(x, y)  # noqa: F821 undefined name
+        self.assertTrue(a is x)  # neither was coerced
         self.assertTrue(b is y)
 
         x = self._makeOne(1)
         y = 2.1
-        a, b = coerce(x, y)
-        self.assertTrue(isinstance(a, float)) # a was coerced
+        a, b = coerce(x, y)  # noqa: F821 undefined name
+        self.assertTrue(isinstance(a, float))  # a was coerced
         self.assertFalse(a is x)
-        self.assertEqual(a,  float(x))
+        self.assertEqual(a, float(x))
         self.assertTrue(b is y)
 
         x = self._makeOne(1.1)
         y = 2
-        a, b = coerce(x, y)
+        a, b = coerce(x, y)  # noqa: F821 undefined name
         self.assertTrue(a is x)
-        self.assertTrue(isinstance(b, float)) # b was coerced
+        self.assertTrue(isinstance(b, float))  # b was coerced
         self.assertFalse(b is y)
-        self.assertEqual(b,  float(y))
+        self.assertEqual(b, float(y))
 
         x = 1
         y = self._makeOne(2)
-        a, b = coerce(x, y)
-        self.assertTrue(a is x) # neither was coerced
+        a, b = coerce(x, y)  # noqa: F821 undefined name
+        self.assertTrue(a is x)  # neither was coerced
         self.assertTrue(b is y)
 
         x = 1.1
         y = self._makeOne(2)
-        a, b = coerce(x, y)
+        a, b = coerce(x, y)  # noqa: F821 undefined name
         self.assertTrue(a is x)
-        self.assertTrue(isinstance(b, float)) # b was coerced
+        self.assertTrue(isinstance(b, float))  # b was coerced
         self.assertFalse(b is y)
-        self.assertEqual(b,  float(y))
+        self.assertEqual(b, float(y))
 
         x = 1
         y = self._makeOne(2.1)
-        a, b = coerce(x, y)
-        self.assertTrue(isinstance(a, float)) # a was coerced
+        a, b = coerce(x, y)  # noqa: F821 undefined name
+        self.assertTrue(isinstance(a, float))  # a was coerced
         self.assertFalse(a is x)
-        self.assertEqual(a,  float(x))
+        self.assertEqual(a, float(x))
         self.assertTrue(b is y)
 
     def test___class__(self):
@@ -650,11 +729,13 @@ class PyProxyBaseTestCase(unittest.TestCase):
         class Descriptor(object):
             value = None
             instance = None
+
             def __set__(self, instance, value):
                 self.value = value
                 self.instance = instance
 
         descriptor = Descriptor()
+
         class Proxy(self._getTargetClass()):
             attr = descriptor
 
@@ -682,6 +763,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
 
         descriptor = Descriptor()
         descriptor.value = "descriptor value"
+
         class Proxy(self._getTargetClass()):
             attr = descriptor
 
@@ -706,14 +788,14 @@ class PyProxyBaseTestCase(unittest.TestCase):
         self.assertRaises(AttributeError, setattr, proxy, 'attr', 42)
         self.assertEqual(proxy.attr, "constant value")
 
-    def _check_wrapping_builtin_returns_correct_provided_by(self, proxy_class, builtin_type):
+    def _check_wrapping_builtin_returns_correct_provided_by(
+            self, proxy_class, builtin_type):
         # We get the __implemented__ (fallback) of the type, not our own
         from zope.interface import Interface
         from zope.interface import classImplements
         from zope.interface import classImplementsOnly
         from zope.interface import implementedBy
         from zope.interface import providedBy
-        from zope.interface import implementedBy
 
         # Set up the builtin interface
         class IFoo(Interface):
@@ -734,38 +816,38 @@ class PyProxyBaseTestCase(unittest.TestCase):
             provided_instance = providedBy(proxy_instance)
             self.assertTrue(IFoo in list(provided_instance))
 
-            # XXX: PyPy 2.5.0 has a bug where proxys around types
-            # aren't correctly hashable, which breaks this part of the
-            # test. This is fixed in 2.5.1, but as of 2015-05-28,
-            # TravisCI still uses 2.5.0.
             proxy_type = proxy_class(builtin_type)
-            from zope.interface.declarations import BuiltinImplementationSpecifications
-            if proxy_type in BuiltinImplementationSpecifications \
-               and BuiltinImplementationSpecifications.get(proxy_type, self) is not self:
-                provided_type = implementedBy(proxy_type)
-                self.assertTrue(IFoo in list(provided_type))
-            else:
-                import sys
-                self.assertEqual((2,5,0), sys.pypy_version_info[:3])
+            from zope.interface.declarations import \
+                BuiltinImplementationSpecifications
+            self.assertIn(proxy_type, BuiltinImplementationSpecifications)
+            self.assertIsNot(
+                BuiltinImplementationSpecifications.get(proxy_type, self),
+                self)
+            provided_type = implementedBy(proxy_type)
+            self.assertTrue(IFoo in list(provided_type))
         finally:
             classImplementsOnly(builtin_type, *impl_before)
 
     def test_wrapping_builtin_type_returns_correct_provided_by(self):
-        self._check_wrapping_builtin_returns_correct_provided_by(self._getTargetClass(), list)
+        self._check_wrapping_builtin_returns_correct_provided_by(
+            self._getTargetClass(), list)
 
-    def _check_wrapping_builtin_with_subclass_returns_correct_provided_by(self, builtin_type):
+    def _check_wrapping_builtin_with_subclass_returns_correct_provided_by(
+            self, builtin_type):
         class Proxy(self._getTargetClass()):
             pass
 
-        self._check_wrapping_builtin_returns_correct_provided_by(Proxy, builtin_type)
+        self._check_wrapping_builtin_returns_correct_provided_by(
+            Proxy, builtin_type)
         # Our new class did not gain an __implemented__ attribute, unless we're
         # the pure-python version
-        if hasattr(Proxy, '__implemented__'):
+        if hasattr(Proxy, '__implemented__'):  # pragma: no cover
             from zope.proxy import PyProxyBase
             self.assertTrue(self._getTargetClass() is PyProxyBase)
 
     def test_wrapping_builtin_with_subclass_returns_correct_provided_by(self):
-        self._check_wrapping_builtin_with_subclass_returns_correct_provided_by(list)
+        self._check_wrapping_builtin_with_subclass_returns_correct_provided_by(
+            list)
 
     def test_method_in_proxy_subclass(self):
         class Proxy(self._getTargetClass()):
@@ -780,15 +862,9 @@ class PyProxyBaseTestCase(unittest.TestCase):
         self.assertNotEqual(getattr(proxy, '__getitem__'), self)
 
     def test_string_to_int(self):
-        # XXX Implementation difference: This works in the
-        # Pure-Python version, but fails in CPython.
-        # See https://github.com/zopefoundation/zope.proxy/issues/4
         proxy = self._makeOne("14")
-        try:
-            self.assertEqual(14, int(proxy))
-        except TypeError:
-            from zope.proxy import PyProxyBase
-            self.assertNotEqual(self._getTargetClass(), PyProxyBase)
+        self.assertEqual(14, int(proxy))
+
 
 class ProxyBaseTestCase(PyProxyBaseTestCase):
 
@@ -796,13 +872,15 @@ class ProxyBaseTestCase(PyProxyBaseTestCase):
         from zope.proxy import ProxyBase
         return ProxyBase
 
+
 class Test_py__module(unittest.TestCase):
     # Historically, proxying __module__ has been troublesome,
     # especially when subclasses of the proxy class are involved;
     # there was also a discrepancy between the C and Python implementations
-    # in that the C implementation only failed Test_subclass__module:test__module__in_instance,
+    # in that the C implementation only failed
+    # Test_subclass__module:test__module__in_instance,
     # whereas the Python version failed every test.
-    # See https://github.com/zopefoundation/zopetoolkit/pull/2#issuecomment-106075153
+    # See https://github.com/zopefoundation/zopetoolkit/pull/2#issuecomment-106075153  # noqa: E501 line too long
     # and https://github.com/zopefoundation/zope.proxy/pull/8
 
     def _getTargetClass(self):
@@ -810,7 +888,6 @@ class Test_py__module(unittest.TestCase):
         return PyProxyBase
 
     def _makeProxy(self, obj):
-        from zope.proxy import PyProxyBase
         return self._getTargetClass()(obj)
 
     def _check_module(self, obj, expected):
@@ -846,6 +923,7 @@ class Test_py__module(unittest.TestCase):
         class Module(object):
             def __init__(self):
                 self.__module__ = __name__
+
             def __eq__(self, other):
                 return self.__module__ == other.__module__
 
@@ -858,23 +936,28 @@ class Test_py__module(unittest.TestCase):
         self.assertEqual(module, self._makeProxy(module))
         self.assertEqual(self._makeProxy(module), module)
 
+
 class Test__module(Test_py__module):
 
     def _getTargetClass(self):
         from zope.proxy import ProxyBase
         return ProxyBase
 
+
 class Test_py_subclass__module(Test_py__module):
 
     def _getTargetClass(self):
-        class ProxySubclass(super(Test_py_subclass__module,self)._getTargetClass()):
+        class ProxySubclass(
+                super(Test_py_subclass__module, self)._getTargetClass()):
             pass
         return ProxySubclass
 
+
 class Test_subclass__module(Test__module):
 
     def _getTargetClass(self):
-        class ProxySubclass(super(Test_subclass__module,self)._getTargetClass()):
+        class ProxySubclass(
+                super(Test_subclass__module, self)._getTargetClass()):
             pass
         return ProxySubclass
 
@@ -910,6 +993,7 @@ class Test_py_getProxiedObject(unittest.TestCase):
         proxy2 = self._makeProxy(proxy)
         self.assertTrue(self._callFUT(proxy2) is proxy)
 
+
 class Test_getProxiedObject(Test_py_getProxiedObject):
 
     def _callFUT(self, *args):
@@ -995,6 +1079,7 @@ class Test_py_isProxy(unittest.TestCase):
     def test_proxy_no_class(self):
         class P1(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1004,6 +1089,7 @@ class Test_py_isProxy(unittest.TestCase):
     def test_proxy_w_same_class(self):
         class P1(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1013,8 +1099,10 @@ class Test_py_isProxy(unittest.TestCase):
     def test_proxy_w_other_class(self):
         class P1(self._proxyClass()):
             pass
+
         class P2(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1043,6 +1131,12 @@ class Test_py_sameProxiedObjects(unittest.TestCase):
         from zope.proxy import PyProxyBase
         return PyProxyBase(obj)
 
+    def _makeSecurityProxy(self, obj):
+        from zope.security.proxy import ProxyPy
+        from zope.security.checker import CheckerPy
+        checker = CheckerPy({})
+        return ProxyPy(obj, checker)
+
     def test_bare_instance_identical(self):
         class C(object):
             pass
@@ -1074,6 +1168,7 @@ class Test_py_sameProxiedObjects(unittest.TestCase):
 
     def test_proxies_w_same_bare(self):
         _mP = self._makeProxy
+
         class C(object):
             pass
         c1 = C()
@@ -1081,6 +1176,7 @@ class Test_py_sameProxiedObjects(unittest.TestCase):
 
     def test_proxies_w_other_bare(self):
         _mP = self._makeProxy
+
         class C(object):
             pass
         c1 = C()
@@ -1090,6 +1186,7 @@ class Test_py_sameProxiedObjects(unittest.TestCase):
 
     def test_nested_proxy_and_same_bare(self):
         _mP = self._makeProxy
+
         class C(object):
             pass
         c1 = C()
@@ -1098,6 +1195,7 @@ class Test_py_sameProxiedObjects(unittest.TestCase):
 
     def test_nested_proxy_and_other_bare(self):
         _mP = self._makeProxy
+
         class C(object):
             pass
         c1 = C()
@@ -1105,6 +1203,21 @@ class Test_py_sameProxiedObjects(unittest.TestCase):
         self.assertFalse(self._callFUT(_mP(_mP(c1)), c2))
         self.assertFalse(self._callFUT(c2, _mP(_mP(c1))))
 
+    @unittest.skipUnless(_HAVE_ZOPE_SECURITY, 'zope.security missing')
+    def test_security_proxy(self):
+        class C(object):
+            pass
+        c1 = C()
+        proxy1 = self._makeSecurityProxy(c1)
+        proxy1_2 = self._makeSecurityProxy(c1)
+
+        self.assertTrue(self._callFUT(proxy1, proxy1))
+        self.assertTrue(self._callFUT(proxy1, proxy1_2))
+
+        c2 = C()
+        proxy2 = self._makeSecurityProxy(c2)
+        self.assertFalse(self._callFUT(proxy1, proxy2))
+
 
 class Test_sameProxiedObjects(Test_py_sameProxiedObjects):
 
@@ -1116,6 +1229,12 @@ class Test_sameProxiedObjects(Test_py_sameProxiedObjects):
         from zope.proxy import ProxyBase
         return ProxyBase(obj)
 
+    def _makeSecurityProxy(self, obj):
+        from zope.security.proxy import Proxy
+        from zope.security.checker import Checker
+        checker = Checker({})
+        return Proxy(obj, checker)
+
 
 class Test_py_queryProxy(unittest.TestCase):
 
@@ -1136,6 +1255,7 @@ class Test_py_queryProxy(unittest.TestCase):
     def test_proxy_no_class(self):
         class P1(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1145,6 +1265,7 @@ class Test_py_queryProxy(unittest.TestCase):
     def test_proxy_w_same_class(self):
         class P1(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1155,8 +1276,10 @@ class Test_py_queryProxy(unittest.TestCase):
     def test_proxy_w_other_class(self):
         class P1(self._proxyClass()):
             pass
+
         class P2(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1167,8 +1290,10 @@ class Test_py_queryProxy(unittest.TestCase):
     def test_proxy_w_base_class(self):
         class P1(self._proxyClass()):
             pass
+
         class P2(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1207,6 +1332,7 @@ class Test_py_queryInnerProxy(unittest.TestCase):
     def test_proxy_no_class(self):
         class P1(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1216,6 +1342,7 @@ class Test_py_queryInnerProxy(unittest.TestCase):
     def test_proxy_w_same_class(self):
         class P1(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1226,8 +1353,10 @@ class Test_py_queryInnerProxy(unittest.TestCase):
     def test_nested_proxy(self):
         class P1(self._proxyClass()):
             pass
+
         class P2(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1241,8 +1370,10 @@ class Test_py_queryInnerProxy(unittest.TestCase):
     def test_re_nested_proxy(self):
         class P1(self._proxyClass()):
             pass
+
         class P2(self._proxyClass()):
             pass
+
         class C(object):
             pass
         c = C()
@@ -1255,7 +1386,7 @@ class Test_py_queryInnerProxy(unittest.TestCase):
         self.assertTrue(self._callFUT(proxy3, P2, 42) is proxy2)
 
 
-class Test_queryInnerProxy(unittest.TestCase):
+class Test_queryInnerProxy(Test_py_queryInnerProxy):
 
     def _callFUT(self, *args):
         from zope.proxy import queryInnerProxy
@@ -1302,14 +1433,16 @@ class Test_py_removeAllProxies(unittest.TestCase):
         proxy2 = self._makeProxy(proxy)
         self.assertTrue(self._callFUT(proxy2) is c)
 
+    @unittest.skipUnless(_HAVE_ZOPE_SECURITY, 'zope.security missing')
     def test_security_proxy(self):
         class C(object):
             pass
         c = C()
         proxy = self._makeSecurityProxy(c)
-        self.assertTrue(self._callFUT(proxy) is c)
+        self.assertIs(self._callFUT(proxy), c)
 
-class Test_removeAllProxies(unittest.TestCase):
+
+class Test_removeAllProxies(Test_py_removeAllProxies):
 
     def _callFUT(self, *args):
         from zope.proxy import removeAllProxies
@@ -1319,6 +1452,11 @@ class Test_removeAllProxies(unittest.TestCase):
         from zope.proxy import ProxyBase
         return ProxyBase(obj)
 
+    def _makeSecurityProxy(self, obj):
+        from zope.security.proxy import Proxy
+        checker = object()
+        return Proxy(obj, checker)
+
 
 class Test_ProxyIterator(unittest.TestCase):
 
@@ -1334,6 +1472,7 @@ class Test_ProxyIterator(unittest.TestCase):
 
     def test_w_simple_proxy(self):
         from zope.proxy import ProxyBase
+
         class C(object):
             pass
         c = C()
@@ -1342,6 +1481,7 @@ class Test_ProxyIterator(unittest.TestCase):
 
     def test_w_nested_proxies(self):
         from zope.proxy import ProxyBase
+
         class C(object):
             pass
         c = C()
@@ -1358,15 +1498,19 @@ class Test_nonOverridable(unittest.TestCase):
     def test_it(self):
         from zope.proxy import ProxyBase
         from zope.proxy import non_overridable
+
         class Proxy(ProxyBase):
             def who(self):
-                return 'PROXY'
+                raise AssertionError("Not called")
+
             @non_overridable
             def what(self):
                 return 'PROXY'
+
         class Foo(object):
             def who(self):
                 return 'FOO'
+
             def what(self):
                 return 'FOO'
         p0 = ProxyBase(Foo())
@@ -1377,6 +1521,32 @@ class Test_nonOverridable(unittest.TestCase):
         self.assertEqual(proxy.what(), 'PROXY')
 
 
+class TestEmptyInterfaceDescriptor(unittest.TestCase):
+
+    def _makeOne(self):
+        from zope.proxy import _EmptyInterfaceDescriptor
+
+        class It(object):
+            feature = _EmptyInterfaceDescriptor()
+        return It()
+
+    def test_set(self):
+        it = self._makeOne()
+        with self.assertRaises(TypeError):
+            it.feature = 42
+
+    def test_delete(self):
+        it = self._makeOne()
+        del it.feature
+        with self.assertRaises(AttributeError):
+            getattr(it, 'feature')
+
+    def test_iter(self):
+        it = type(self._makeOne())
+        feature = it.__dict__['feature']
+        self.assertEqual([], list(feature))
+
+
 class Comparable(object):
     def __init__(self, value):
         self.value = value
@@ -1399,33 +1569,9 @@ class Comparable(object):
     def __gt__(self, other):
         return not self.__le__(other)
 
-    def __repr__(self):
+    def __repr__(self):  # pragma: no cover
         return "<Comparable: %r>" % self.value
 
 
 def test_suite():
-    return unittest.TestSuite((
-        unittest.makeSuite(ModuleConformanceCase),
-        unittest.makeSuite(PyProxyBaseTestCase),
-        unittest.makeSuite(ProxyBaseTestCase),
-        unittest.makeSuite(Test_py_getProxiedObject),
-        unittest.makeSuite(Test_getProxiedObject),
-        unittest.makeSuite(Test_py_setProxiedObject),
-        unittest.makeSuite(Test_setProxiedObject),
-        unittest.makeSuite(Test_py_isProxy),
-        unittest.makeSuite(Test_isProxy),
-        unittest.makeSuite(Test_py_sameProxiedObjects),
-        unittest.makeSuite(Test_sameProxiedObjects),
-        unittest.makeSuite(Test_py_queryProxy),
-        unittest.makeSuite(Test_queryProxy),
-        unittest.makeSuite(Test_py_queryInnerProxy),
-        unittest.makeSuite(Test_queryInnerProxy),
-        unittest.makeSuite(Test_py_removeAllProxies),
-        unittest.makeSuite(Test_removeAllProxies),
-        unittest.makeSuite(Test_ProxyIterator),
-        unittest.makeSuite(Test_nonOverridable),
-        unittest.makeSuite(Test_py__module),
-        unittest.makeSuite(Test__module),
-        unittest.makeSuite(Test_py_subclass__module),
-        unittest.makeSuite(Test_subclass__module),
-    ))
+    return unittest.defaultTestLoader.loadTestsFromName(__name__)
